import json, uuid, logging, thread, time, os, sys, traceback

from flask import current_app

from .slack import SlackMessage

log = logging.getLogger(__name__)

TASKS = {}

def task(*args, **kwargs):
    def deco(f):
        task = Task(current_app.fanny.redis, f.__name__, f, *args, **kwargs)

        if f.__name__ in TASKS:
            raise Exception("Conflicting task name: %s" % f.__name__)

        TASKS[f.__name__] = task
        return task
    return deco

class Task(object):
    def __init__(self, name, f, max_running=None, buffer_time=None, max_size=25):
        self.redis = current_app.fanny.redis
        self.name = name
        self.f = f
        self.max_running = max_running
        self.max_size = max_size
        self.buffer_time = buffer_time

    def __call__(self, *args, **kwargs):
        return self.f(*args, **kwargs)

    def queue(self, *args, **kwargs):
        # Make sure we have space
        if (self.redis.llen("jq:%s" % self.name) or 0) > self.max_size:
            raise Exception("Queue for task %s is full!" % self.name)

        id = str(uuid.uuid4())
        self.redis.rpush("jq:%s" % self.name, json.dumps({
            "id": id,
            "args": args,
            "kwargs": kwargs
        }))
        return id

class TaskRunner(object):
    def __init__(self, name, task):
        self.redis = current_app.fanny.redis
        self.name = name
        self.f = task
        self.running = 0

    def process(self, job):
        self.running += 1
        log.info('[%s] Running job %s...', job['id'], self.name)
        self.redis.set("task:%s" % job['id'], 1)
        start = time.time()

        try:
            self.f(*job['args'], **job['kwargs'])
            if self.f.buffer_time:
                time.sleep(self.f.buffer_time)
        except:
            log.exception("[%s] Failed in %ss", job['id'], time.time() - start)

            # Send slack stack trace
            try:
                exc_type, exc_obj, exc_tb = sys.exc_info()

                content = ''.join(traceback.format_exception(exc_type, exc_obj, exc_tb))
                msg = SlackMessage("Task Exception (%s)" % str(exc_obj), color='danger')
                msg.add_custom_field("Task Name", self.name)
                msg.add_custom_field("Task ID", job['id'])
                msg.add_custom_field("Duration", "%ss" % (time.time() - start))
                msg.add_custom_field("Exception", content)
                msg.send_async()
            except Exception:
                log.exception("Failed to send slack exception trace: ")
        finally:
            self.redis.delete("task:%s" % job['id'])
            self.running -= 1
        log.info('[%s] Completed in %ss', job['id'], time.time() - start)

    def run(self, job):
        if self.f.max_running:
            while self.f.max_running <= self.running:
                time.sleep(.5)

        thread.start_new_thread(self.process, (job, ))

class TaskManager(object):
    def __init__(self, redis):
        self.redis = redis
        self.load()
        self.queues = ["jq:" + i for i in TASKS.keys()]
        self.runners = {k: TaskRunner(self.redis, k, v) for k, v in TASKS.items()}
        self.active = True

    def load(self):
        for f in os.listdir("tasks/"):
            if f.endswith(".py"):
                __import__("tasks." + f.rsplit(".")[0])

    def run(self):
        log.info("Running TaskManager on %s queues...", len(self.queues))
        while self.active:
            chan, job = self.redis.blpop(self.queues)
            job_name = chan.split(":", 1)[1]
            job = json.loads(job)

            if job_name not in TASKS:
                log.error("Cannot handle task %s",job_name)
                continue

            self.runners[job_name].run(job)




from flask import flash, redirect

def flashy(msg, etype="danger", path="/"):
    """
    Flashes a message and redirects a user. Returns a flask Response that can
    be returned to the user.
    """
    flash(msg, etype)
    return redirect(path)




class BaseEnum(object):
    pass

class EnumAttr(object):
    def __init__(self, parent, content, index):
        self.parent = parent
        self.content = content
        self.index = index

    def __eq__(self, other):
        if isinstance(other, EnumAttr):
            (self.index == other.index) or (self.content == other.content)

        if isinstance(other, int):
            if self.index == other:
                return True

        return self.content == other

    def __cmp__(self, other):
        if isinstance(other, EnumAttr):
            return self.index - other.index

        if isinstance(other, int):
            return self.index - other

        return self.index - self.parent.ORDER.index(other)

    def __repr__(self):
        return '<EnumAttr {}>'.format(self.content)

    def __str__(self):
        return self.content

    def __int__(self):
        return self.index

class BaseEnumMeta(type):
    def __getattr__(self, attr):
        if attr.lower() in self.attrs:
            return self.attrs[attr.lower()]
        raise AttributeError

    def __getitem__(self, item):
        if isinstance(item, str):
            return getattr(self, item)
        else:
            return self.attrs[self.order[item]]

def Enum(*args):
    class _T(BaseEnum):
        __metaclass__ = BaseEnumMeta

    _T.order = args
    _T.attrs = {}

    for index, entry in enumerate(args):
        _T.attrs[entry] = EnumAttr(_T, entry, index)
        # setattr(_T, entry, EnumAttr(_T, entry, index))
        # setattr(_T, entry.upper(), getattr(_T, entry))

    return _T




import sys, os

from flask import current_app
from redis import Redis

from .filters import register_filters
from .log import setup_logging
from .session import SessionProvider

class Holster(object):
    def __init__(self, app=None, auto_load_views=True):
        self.app = app or current_app
        self.redis = None

        self.init_app(self.app)

        if auto_load_views:
            self.load_views()

        if "REDIS" in self.app.config:
            self.init_redis(self.app.config.get("REDIS"))
            self.sessions = SessionProvider(self.redis)
            self.app.sessions = self.sessions

    def init_redis(self, kwargs):
        self.redis = Redis(**kwargs)

    def init_app(self, app):
        app.holster = self
        register_filters(app)
        setup_logging(app)

    def load_views(self):
        for view in os.listdir("views"):
            if not view.endswith(".py") or view.startswith("_"):
                continue

            view = "views." + view.split(".py")[0]
            __import__(view)
            self.app.register_blueprint(getattr(sys.modules[view], view.split(".")[-1]))




import json, logging
from slacker import Slacker

from flask import current_app

log = logging.getLogger(__name__)
slack = Slacker('xoxb-3809520008-UCTKTT9r8AKOLUYXHhsk5jER')

class SlackMessage(object):
    def __init__(self, content, color=None, username="empburt", channel="#ops"):
        self.text = content
        self.color = color
        self.username = username
        self.channel = channel

        self.fallback = content
        self.pretext = None
        self.author_name = None
        self.author_link = None
        self.author_icon = None
        self.title = None
        self.title_link = None
        self.image_url = None

        self.fields = []

    def add_custom_field(self, title, value, short=None):
        short = len(str(value)) < 64 if short is None else short

        self.fields.append({
            "title": title,
            "value": value,
            "short": short
        })

    def payload(self):
        return (self.channel, ""), {
            'username': self.username,
            'attachments': [{
                "fallback": self.fallback,
                "color": self.color,
                "pretext": self.pretext,
                "author_name": self.author_name,
                "author_link": self.author_link,
                "author_icon": self.author_icon,
                "title": self.title,
                "title_link": self.title_link,
                "text": self.text,
                "fields": self.fields,
                "image_url": self.image_url
            }]}

    def send_async(self):
        if current_app.config.get("SLACK_DISABLE"):
            log.debug("Would send slack message: %s, %s, %s, %s" % (
                self.text, self.color, self.fields, self.username))
            return

        from tasks.slack import slack_async_message
        args, kwargs = self.payload()
        slack_async_message.queue({"args": args, "kwargs": kwargs})

    def send(self):
        if current_app.config.get("SLACK_DISABLE"):
            log.debug("Would send slack message: %s, %s, %s, %s" % (
                self.text, self.color, self.fields, self.username))
            return
        self.send_raw(*self.payload())

    @staticmethod
    def send_raw(args, kwargs):
        kwargs['attachments'] = json.dumps(kwargs['attachments'])
        return slack.chat.post_message(*args, **kwargs)



import mandrill, logging

from flask import current_app

log = logging.getLogger(__name__)
mandrill_client = mandrill.Mandrill(current_app.config.get("EMAIL_MANDRILL_KEY"))

class Email(object):
    def __init__(self):
        self.from_addr = "noreply@csgofort.com"
        self.from_name = "CSGO Fort"
        self.to_addrs = []
        self.subject = ""
        self.body = ""

    def send(self):
        payload = {
            "from_email": self.from_addr,
            "from_name": self.from_name,
            "html": self.body,
            "subject": self.subject,
            "to": map(lambda i: {
                "email": i, "type": "to"
            }, self.to_addrs),
        }

        if current_app.config.get("EMAIL_DISABLE"):
            log.debug("Would send email %s" % payload)
            return

        log.info("Sending email to %s" % self.to_addrs)
        mandrill_client.messages.send(message=payload)

